@@ -5,6 +5,7 @@ import asyncio
 import datetime
 import json
 import logging
+import re
 from typing import Any, Callable, Dict, Union
 
 import aiohttp  # type: ignore
@@ -12,10 +13,23 @@ from aiohttp.client_exceptions import ContentTypeError, ServerTimeoutError
 from awesomeversion import AwesomeVersion
 from awesomeversion.exceptions import AwesomeVersionCompareException
 
-from .const import MAX_AMPS, MIN_AMPS
+from .const import (
+    BAT_LVL,
+    BAT_RANGE,
+    GRID,
+    MAX_AMPS,
+    MIN_AMPS,
+    RELEASE,
+    SOLAR,
+    TTF,
+    TYPE,
+    VALUE,
+    VOLTAGE,
+)
 from .exceptions import (
     AlreadyListening,
     AuthenticationError,
+    InvalidType,
     MissingMethod,
     MissingSerial,
     ParseJSONError,
@@ -120,13 +134,17 @@ class OpenEVSE:
                         _LOGGER.warning("Non JSON response: %s", message)
 
                     if resp.status == 400:
-                        _LOGGER.error("Error 400: %s", message["msg"])
+                        if "msg" in message.keys():
+                            index = "msg"
+                        elif "error" in message.keys():
+                            index = "error"
+                        _LOGGER.error("Error 400: %s", message[index])
                         raise ParseJSONError
                     if resp.status == 401:
                         _LOGGER.error("Authentication error: %s", message)
                         raise AuthenticationError
                     if resp.status in [404, 405, 500]:
-                        _LOGGER.error("%s", message)
+                        _LOGGER.warning("%s", message)
 
                     if method == "post" and "config_version" in message:
                         await self.update()
@@ -552,7 +570,16 @@ class OpenEVSE:
         if max_version != "":
             limit = AwesomeVersion(max_version)
 
+        firmware_filtered = None
+
+        try:
+            firmware_search = re.search("\\d\\.\\d\\.\\d", self._config["version"])
+            if firmware_search is not None:
+                firmware_filtered = firmware_search[0]
+        except Exception:  # pylint: disable=broad-exception-caught
+            _LOGGER.warning("Non-standard versioning string.")
         _LOGGER.debug("Detected firmware: %s", self._config["version"])
+        _LOGGER.debug("Filtered firmware: %s", firmware_filtered)
 
         if "dev" in self._config["version"]:
             value = self._config["version"]
@@ -562,7 +589,7 @@ class OpenEVSE:
         elif "master" in self._config["version"]:
             value = "dev"
         else:
-            value = self._config["version"]
+            value = firmware_filtered
 
         current = AwesomeVersion(value)
 
@@ -581,10 +608,35 @@ class OpenEVSE:
             _LOGGER.debug("Non-semver firmware version detected.")
         return False
 
+    # HTTP Posting of grid voltage
+
+    async def grid_voltage(self, voltage: int | None = None) -> None:
+        """Send pushed sensor data to grid voltage."""
+        if not self._version_check("4.0.0"):
+            _LOGGER.debug("Feature not supported for older firmware.")
+            raise UnsupportedFeature
+
+        url = f"{self.url}status"
+        data = {}
+
+        if voltage is not None:
+            data[VOLTAGE] = voltage
+
+        if not data:
+            _LOGGER.info("No sensor data to send to device.")
+        else:
+            _LOGGER.debug("Posting voltage: %s", data)
+            response = await self.process_request(url=url, method="post", data=data)
+            _LOGGER.debug("Voltage posting response: %s", response)
+
     # Self production HTTP Posting
 
     async def self_production(
-        self, grid: int | None = None, solar: int | None = None, invert: bool = True
+        self,
+        grid: int | None = None,
+        solar: int | None = None,
+        invert: bool = True,
+        voltage: int | None = None,
     ) -> None:
         """Send pushed sensor data to self-prodcution."""
         if not self._version_check("4.0.0"):
@@ -600,9 +652,11 @@ class OpenEVSE:
 
         # Prefer grid sensor data
         if grid is not None:
-            data = {"grid_ie": grid}
+            data[GRID] = grid
         elif solar is not None:
-            data = {"solar": solar}
+            data[SOLAR] = solar
+        if voltage is not None:
+            data[VOLTAGE] = voltage
 
         if not data:
             _LOGGER.info("No sensor data to send to device.")
@@ -611,6 +665,97 @@ class OpenEVSE:
             response = await self.process_request(url=url, method="post", data=data)
             _LOGGER.debug("Self-production response: %s", response)
 
+    # State of charge HTTP posting
+    async def soc(
+        self,
+        battery_level: int | None = None,
+        battery_range: int | None = None,
+        time_to_full: int | None = None,
+        voltage: int | None = None,
+    ) -> None:
+        """Send pushed sensor data to self-prodcution."""
+        if not self._version_check("4.1.0"):
+            _LOGGER.debug("Feature not supported for older firmware.")
+            raise UnsupportedFeature
+
+        url = f"{self.url}status"
+        data = {}
+
+        # Build post data
+        if battery_level is not None:
+            data[BAT_LVL] = battery_level
+        if battery_range is not None:
+            data[BAT_RANGE] = battery_range
+        if time_to_full is not None:
+            data[TTF] = time_to_full
+        if voltage is not None:
+            data[VOLTAGE] = voltage
+
+        if not data:
+            _LOGGER.info("No SOC data to send to device.")
+        else:
+            _LOGGER.debug("Posting SOC data: %s", data)
+            response = await self.process_request(url=url, method="post", data=data)
+            _LOGGER.debug("SOC response: %s", response)
+
+    # Limit endpoint
+    async def set_limit(
+        self, limit_type: str, value: int, release: bool | None = None
+    ) -> Any:
+        """Set charge limit."""
+        if not self._version_check("5.0.0"):
+            _LOGGER.debug("Feature not supported for older firmware.")
+            raise UnsupportedFeature
+
+        url = f"{self.url}limit"
+        data: Dict[str, Any] = {}
+        valid_types = ["time", "energy", "soc", "range"]
+
+        if limit_type not in valid_types:
+            raise InvalidType
+
+        data[TYPE] = limit_type
+        data[VALUE] = value
+        if release is not None:
+            data[RELEASE] = release
+
+        _LOGGER.debug("Limit data: %s", data)
+        _LOGGER.debug("Setting limit config on %s", url)
+        response = await self.process_request(
+            url=url, method="post", data=data
+        )  # noqa: E501
+        return response
+
+    async def clear_limit(self) -> Any:
+        """Clear charge limit."""
+        if not self._version_check("5.0.0"):
+            _LOGGER.debug("Feature not supported for older firmware.")
+            raise UnsupportedFeature
+
+        url = f"{self.url}limit"
+        data: Dict[str, Any] = {}
+
+        _LOGGER.debug("Clearing limit config on %s", url)
+        response = await self.process_request(
+            url=url, method="delete", data=data
+        )  # noqa: E501
+        return response
+
+    async def get_limit(self) -> Any:
+        """Get charge limit."""
+        if not self._version_check("5.0.0"):
+            _LOGGER.debug("Feature not supported for older firmware.")
+            raise UnsupportedFeature
+
+        url = f"{self.url}limit"
+        data: Dict[str, Any] = {}
+
+        _LOGGER.debug("Getting limit config on %s", url)
+        response = await self.process_request(
+            url=url, method="get", data=data
+        )  # noqa: E501
+        return response
+
     @property
     def hostname(self) -> str:
         """Return charger hostname."""
